
//		ByArgs.cpp - keystrokes to move and delete by args, rather than words


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				0
#define VER_MINOR				92
#define VER_BUILD				0
#define VER_PATCH				5

#define PLUGIN_NAME				L"ByArgs"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif


#define MAX_ARG_LEN				8192
#define MAX_ARGS				1024
#define SHORT_BUF_LEN			1024


#define MAX_NESTING				512

#define QUOTE					0x22

#define KEY_NONE				0xffff0000

#define LAST_BEFORE_CURSOR		0x00
#define FIRST_AFTER_CURSOR		0x01
#define STOP_AT_CLOSE_BRACKET	0x80


LPTSTR arg[MAX_ARGS];
int arglen[MAX_ARGS];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN]	= L"";



const wchar_t PluginFeatures[] = L"byargs,*key";


const wchar_t HelpWindowTitle[] = L"ByArgs Plugin Help";


const wchar_t ByArgsHelpText[] = L"Display or change settings for the ByArgs plugin.\n\
\n\
BYARGS /ML:key /MR:key /DL:key /DR:key /L /R /S\r\n\
\n\
\t/ML:  set the hotkey(s) to Move Left  one arg (Control-Left)\n\
\t/MR:  set the hotkey(s) to Move Right one arg (Control-Right)\n\
\t/DL:  set the hotkey(s) to Delete Left  one arg (Control-L)\n\
\t/DR:  set the hotkey(s) to Delete Right one arg (Ctrl-R, Ctrl-BkSp)  \n\
\t/L    List all supported hotkeys\n\
\t/R    Restore saved settings from the registry\n\
\t/S    Save settings to the registry as new default\n\
\n\
Examples of hotkeys are Alt-L, Control-Alt-Left, or Alt-[.  To\n\
disable a function, specify NONE as the hotkey.  You can set two\n\
hotkeys per function; separate them with a comma, for example\n\
/DR:CONTROL-ALT-R,CONTROL-ALT-BKSP.\n\
\n\
If no hotkeys are specified, the current settings will be shown.\n";


#include "HotKeys.cpp"



const wchar_t PluginRegKey[]			= L"Software\\JPPlugins\\ByArgs";
const wchar_t MoveLeftRegValue[]		= L"MoveLeft";
const wchar_t MoveRightRegValue[]		= L"MoveRight";
const wchar_t DelLeftRegValue[]			= L"DeleteLeft";
const wchar_t DelRightRegValue[]		= L"DeleteRight";
const wchar_t MoveLeft2RegValue[]		= L"MoveLeft2";
const wchar_t MoveRight2RegValue[]		= L"MoveRight2";
const wchar_t DelLeft2RegValue[]		= L"DeleteLeft2";
const wchar_t DelRight2RegValue[]		= L"DeleteRight2";


DWORD MoveLeftHotKey	= 0x10073;
DWORD MoveRightHotKey	= 0x10074;
DWORD DelLeftHotKey		= 0x0000c;
DWORD DelRightHotKey	= 0x00012;
DWORD MoveLeft2HotKey	= KEY_NONE;
DWORD MoveRight2HotKey	= KEY_NONE;
DWORD DelLeft2HotKey	= KEY_NONE;
DWORD DelRight2HotKey	= 0x0007f;



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), shortbuf );

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/byargs.html";
	piInfo.pszDescription	= (LPTSTR) L"Move or delete by arguments";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



int SaveOptions( BOOL ComplainOnError )
{
	int rv = 0;
	HKEY KeyHandle = NULL;

	if ( RegCreateKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, NULL, 0, KEY_WRITE, NULL, &KeyHandle, NULL ) == ERROR_SUCCESS ) {

		if ( RegSetValueEx( KeyHandle, MoveLeftRegValue, 0, REG_DWORD, (LPBYTE) &MoveLeftHotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, MoveRightRegValue, 0, REG_DWORD, (LPBYTE) &MoveRightHotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, DelLeftRegValue, 0, REG_DWORD, (LPBYTE) &DelLeftHotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, DelRightRegValue, 0, REG_DWORD, (LPBYTE) &DelRightHotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, MoveLeft2RegValue, 0, REG_DWORD, (LPBYTE) &MoveLeft2HotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, MoveRight2RegValue, 0, REG_DWORD, (LPBYTE) &MoveRight2HotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, DelLeft2RegValue, 0, REG_DWORD, (LPBYTE) &DelLeft2HotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( RegSetValueEx( KeyHandle, DelRight2RegValue, 0, REG_DWORD, (LPBYTE) &DelRight2HotKey, sizeof( DWORD ) ) != ERROR_SUCCESS )
			rv = -1;

		if ( rv && ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Error writing to registry\r\n" );
		}
		rv = -1;

		RegCloseKey( KeyHandle );
	}
	else {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Error opening registry key\r\n" );
		}
		rv = -1;
	}

	return rv;
}



int RestoreOptions( BOOL ComplainOnError )
{
	int rv = 0;
	HKEY KeyHandle = NULL;
	DWORD mlval = 0xffff0000, mrval = 0xffff0000, dlval = 0xffff0000, drval = 0xffff0000;
	DWORD ml2val = 0xffff0000, mr2val = 0xffff0000, dl2val = 0xffff0000, dr2val = 0xffff0000;
	DWORD buflen = 0;
	

	if ( RegOpenKeyEx( HKEY_CURRENT_USER, PluginRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
		rv = 0;
		int k = 0;

		buflen = sizeof( DWORD );
		if ( RegQueryValueEx( KeyHandle, MoveLeftRegValue, NULL, NULL, (LPBYTE) &mlval, &buflen ) == ERROR_SUCCESS )
			k++;

		buflen = sizeof( DWORD );
		if ( RegQueryValueEx( KeyHandle, MoveRightRegValue, NULL, NULL, (LPBYTE) &mrval, &buflen ) == ERROR_SUCCESS )
			k++;

		buflen = sizeof( DWORD );
		if ( RegQueryValueEx( KeyHandle, DelLeftRegValue, NULL, NULL, (LPBYTE) &dlval, &buflen ) == ERROR_SUCCESS )
			k++;

		buflen = sizeof( DWORD );
		if ( RegQueryValueEx( KeyHandle, DelRightRegValue, NULL, NULL, (LPBYTE) &drval, &buflen ) == ERROR_SUCCESS )
			k++;

		buflen = sizeof( DWORD );
		RegQueryValueEx( KeyHandle, MoveLeft2RegValue, NULL, NULL, (LPBYTE) &ml2val, &buflen );

		buflen = sizeof( DWORD );
		RegQueryValueEx( KeyHandle, MoveRight2RegValue, NULL, NULL, (LPBYTE) &mr2val, &buflen );

		buflen = sizeof( DWORD );
		RegQueryValueEx( KeyHandle, DelLeft2RegValue, NULL, NULL, (LPBYTE) &dl2val, &buflen );

		buflen = sizeof( DWORD );
		RegQueryValueEx( KeyHandle, DelRight2RegValue, NULL, NULL, (LPBYTE) &dr2val, &buflen );

		RegCloseKey( KeyHandle );

		if ( !k )
			rv = -1;
	}
	else
		rv = -1;

	if ( rv ) {
		if ( RegOpenKeyEx( HKEY_LOCAL_MACHINE, PluginRegKey, 0, KEY_READ, &KeyHandle ) == ERROR_SUCCESS ) {
			rv = 0;
			int k = 0;

			buflen = sizeof( DWORD );
			if ( RegQueryValueEx( KeyHandle, MoveLeftRegValue, NULL, NULL, (LPBYTE) &mlval, &buflen ) == ERROR_SUCCESS )
				k++;

			buflen = sizeof( DWORD );
			if ( RegQueryValueEx( KeyHandle, MoveRightRegValue, NULL, NULL, (LPBYTE) &mrval, &buflen ) == ERROR_SUCCESS )
				k++;

			buflen = sizeof( DWORD );
			if ( RegQueryValueEx( KeyHandle, DelLeftRegValue, NULL, NULL, (LPBYTE) &dlval, &buflen ) == ERROR_SUCCESS )
				k++;

			buflen = sizeof( DWORD );
			if ( RegQueryValueEx( KeyHandle, DelRightRegValue, NULL, NULL, (LPBYTE) &drval, &buflen ) == ERROR_SUCCESS )
				k++;

			buflen = sizeof( DWORD );
			RegQueryValueEx( KeyHandle, MoveLeft2RegValue, NULL, NULL, (LPBYTE) &ml2val, &buflen );

			buflen = sizeof( DWORD );
			RegQueryValueEx( KeyHandle, MoveRight2RegValue, NULL, NULL, (LPBYTE) &mr2val, &buflen );

			buflen = sizeof( DWORD );
			RegQueryValueEx( KeyHandle, DelLeft2RegValue, NULL, NULL, (LPBYTE) &dl2val, &buflen );

			buflen = sizeof( DWORD );
			RegQueryValueEx( KeyHandle, DelRight2RegValue, NULL, NULL, (LPBYTE) &dr2val, &buflen );

			RegCloseKey( KeyHandle );

			if ( !k )
				rv = -1;
		}
	}

	if ( rv ) {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Error reading from registry\r\n" );
		}

		return rv;
	}

	/*
	rv = -2;

	for ( int i = 0; i < KEY_DEFINITIONS; i++ )
		for ( int j = 0; j < 6; j++ )
			if ( keyval == KeyValue[i][j] )
				rv = 0;

	if ( rv < 0 ) {
		if ( ComplainOnError ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Invalid hotkey value in registry\r\n" );
		}
		return rv;
	}
	*/

	MoveLeftHotKey		= mlval;
	MoveRightHotKey		= mrval;
	DelLeftHotKey		= dlval;
	DelRightHotKey		= drval;
	MoveLeft2HotKey		= ml2val;
	MoveRight2HotKey	= mr2val;
	DelLeft2HotKey		= dl2val;
	DelRight2HotKey		= dr2val;
	return 0;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	InitHotkeys();

	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";

	if ( RestoreOptions( FALSE ) ) {
		MoveLeftHotKey		= 0x10073;
		MoveRightHotKey		= 0x10074;
		DelLeftHotKey		= 0x0000c;
		DelRightHotKey		= 0x00012;
		MoveLeft2HotKey		= KEY_NONE;
		MoveRight2HotKey	= KEY_NONE;
		DelLeft2HotKey		= KEY_NONE;
		DelRight2HotKey		= 0x0007f;
	}

	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	return 0;
}



#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_EQUALS_KLUDGE		16
#define PARSE_ONE_ARG_KLUDGE	32
#define PARSE_RETAIN_QUOTES		64
#define PARSE_FORCE_UPPERCASE	128
#define PARSE_NO_TRIM			256
#define PARSE_IGNORE_QUOTES		512
#define PARSE_EQ_QT_KLUDGE		1024


int ParseArgs( LPTSTR inString, unsigned int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 4 - equals kludge:    break at first unquoted equals sign
				bit 5 - one-arg kludge:   if arg does not begin with /, allow unquoted spaces
				bit 6 - don't swallow double quotes
				bit 7 - force uppercase
				bit 8 - don't trim spaces from the right
				bit 9 - disable all special handling of double quotes
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;
	int eqkl    = 0;

	if ( ( flags & ( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS | PARSE_EQUALS_KLUDGE ) ) == 0 )
		flags |= PARSE_BREAK_SPACES;

	for ( i = 0; i < MAX_ARGS; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ( ch == '"' ) && ( ( flags & PARSE_IGNORE_QUOTES ) == 0 ) ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & PARSE_QUOTES_KLUDGE ) ) {
				if ( flags & PARSE_RETAIN_QUOTES ) {
					if ( b < ( MAX_ARG_LEN - 1 ) )
						allargs[b++] = ':';
				}
				else
					ch = ':';
			}
			else if ( ( flags & PARSE_RETAIN_QUOTES ) == 0 )
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) && !quote )
			continue;

		if ( ( a == 0 ) && ( ch != '/' ) && ( ( flags & ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) == ( PARSE_BREAK_SPACES | PARSE_ONE_ARG_KLUDGE ) ) )
			flags &= ~PARSE_BREAK_SPACES;

		if ( iswspace( ch ) && !quote && ( a == 1 ) && ( argnum > 0 ) && ( flags & PARSE_EQUALS_KLUDGE ) && ( arg[argnum][0] == '=' ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & PARSE_BREAK_SPACES ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & PARSE_BREAK_COMMAS ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & PARSE_SLASHES_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ( ch == '=' ) && !quote && ( eqkl == 0 ) && ( a > 0 ) && ( arg[argnum][0] != '/' ) && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			eqkl = argnum;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );
			//- Printf( L"   * equals kludge invoked; eqkl = %d\n", eqkl );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}
		else if ( ( ch == '=' ) && !quote && ( flags & PARSE_EQUALS_KLUDGE ) ) {
			eqkl = -1;
			flags &= ~( PARSE_BREAK_SPACES | PARSE_BREAK_COMMAS );

			if ( flags & PARSE_EQ_QT_KLUDGE )
				flags |= PARSE_IGNORE_QUOTES;
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			quote = FALSE;
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Buffer overflow: b = %d, MAX_ARG_LEN = %d \r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARGS ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & PARSE_FORCE_UPPERCASE )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	if ( ( flags & PARSE_NO_TRIM ) == 0 ) {
		for ( i = 0; i < argnum; i++ ) {
			while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
				arglen[i]--;
				arg[i][arglen[i]] = '\0';
			}
		}
	}

	numargs = argnum;
	return argnum;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



BOOL IsListSep( wchar_t ch )
{
	return ( ( ch == ',' ) || ( ch == ';' ) );
}



BOOL IsVarnameChar( wchar_t ch )
{
	return ( iswalnum( ch ) || ( ch == '_' ) || ( ch == '$' ) );
}



int ParseTwoKeys( LPTSTR inString, DWORD *Key1, DWORD *Key2 )
{
	if ( !inString )
		return -1;
	else if ( inString[0] == '\0' )
		return 1;

	if ( !Key1 || !Key2 )
		return -1;


	wchar_t temp[SHORT_BUF_LEN] = L"";
	DWORD Hotkey1 = KEY_NONE, Hotkey2 = KEY_NONE;

	int i = 0, j = 0;
	wchar_t ch = 0x00, lastch = 0x00;

	do {
		lastch = ch;
		ch = inString[i++];
		if ( ( ch == '\0' ) || ( IsListSep( ch ) && ( lastch != '-' ) ) )
			break;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			temp[j++] = ch;
			temp[j] = '\0';
		}
	} while ( TRUE );

	if ( !j )
		return 1;

	Hotkey1 = ParseKeyName( temp );
	switch ( Hotkey1 ) {
		case 0 :
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad hotkey \"%s\"\r\n", temp );
			return 1;

		case -1 :
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid hotkey \"%s\"\r\n", temp );
			return 1;

		case -2 :
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Key not allowed \"%s\"\r\n", temp );
			return 1;
	}

	if ( IsListSep( ch ) ) {
		if ( inString[i] == '\0' )
			return 1;

		Hotkey2 = ParseKeyName( inString + i );
		switch ( Hotkey2 ) {
			case 0 :
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Bad hotkey \"%s\"\r\n", inString + i );
				return 1;

			case -1 :
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Invalid hotkey \"%s\"\r\n", inString + i );
				return 1;

			case -2 :
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Key not allowed \"%s\"\r\n", inString + i );
				return 1;
		}
	}

	if ( ( Hotkey1 == KEY_NONE) && ( Hotkey2 != KEY_NONE ) ) {
		DWORD t = Hotkey1;
		Hotkey1 = Hotkey2;
		Hotkey2 = t;
	}

	if ( Hotkey1 == Hotkey2 )
		Hotkey2 = KEY_NONE;

	*Key1 = Hotkey1;
	*Key2 = Hotkey2;
	return 0;
}



/*
int PadRight( LPTSTR String, int Len )
{
	if ( !String )
		return -1;

	int l = (int) wcslen( String );
	if ( l >= Len )
		return 0;

	for ( int i = 0; i < ( Len - l ); i++ )
		String[l + i] = 0x20;

	String[Len] = 0x00;
	return 0;
}
*/



int ShowKeyAssignments()
{
	wchar_t KeyName1[SHORT_BUF_LEN]	= L"";
	wchar_t KeyName2[SHORT_BUF_LEN]	= L"";

	GetKeyName( MoveLeftHotKey, KeyName1 );
	GetKeyName( MoveLeft2HotKey, KeyName2 );
	if ( MoveLeft2HotKey == KEY_NONE )
		Printf( L"Move arg left   :  %s\r\n", KeyName1 );
	else
		Printf( L"Move arg left   :  %s or %s\r\n", KeyName1, KeyName2 );

	GetKeyName( MoveRightHotKey, KeyName1 );
	GetKeyName( MoveRight2HotKey, KeyName2 );
	if ( MoveRight2HotKey == KEY_NONE )
		Printf( L"Move arg right  :  %s\r\n", KeyName1 );
	else
		Printf( L"Move arg right  :  %s or %s\r\n", KeyName1, KeyName2 );

	GetKeyName( DelLeftHotKey, KeyName1 );
	if ( DelLeft2HotKey == KEY_NONE )
		Printf( L"Delete arg left :  %s\r\n", KeyName1 );
	else {
		GetKeyName( DelLeft2HotKey, KeyName2 );
		Printf( L"Delete arg left :  %s or %s\r\n", KeyName1, KeyName2 );
	}

	GetKeyName( DelRightHotKey, KeyName1 );
	if ( DelRight2HotKey == KEY_NONE )
		Printf( L"Delete arg right:  %s\r\n", KeyName1 );
	else {
		GetKeyName( DelRight2HotKey, KeyName2 );
		Printf( L"Delete arg right:  %s or %s\r\n", KeyName1, KeyName2 );
	}
	
	/*
	GetKeyName( MoveLeftHotKey, KeyName1 );
	GetKeyName( MoveRightHotKey, KeyName2 );
	PadRight( KeyName1, 20 );
	Printf( L"Move arg left  :  %s  Move arg right  :  %s\r\n", KeyName1, KeyName2 );

	GetKeyName( DelLeftHotKey, KeyName1 );
	GetKeyName( DelRightHotKey, KeyName2 );
	PadRight( KeyName1, 20 );
	Printf( L"Delete arg left:  %s  Delete arg right:  %s\r\n", KeyName1, KeyName2 );
	*/

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI byargs( LPTSTR inString )
{
	if ( !inString )
		return -1;

	BOOL Save = FALSE, Restore = FALSE, List = FALSE, AnyKey = FALSE;
	int rv = 0;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_FORCE_UPPERCASE );

	for ( int i = 0; i < MAX_ARGS; i++ ) {
		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ByArgsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ByArgsHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/S" ) )
				Save = TRUE;
			else if ( !_wcsicmp( arg[i], L"/R" ) )
				Restore = TRUE;
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				List = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/ML", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseTwoKeys( arg[i] + j, &MoveLeftHotKey, &MoveLeft2HotKey ) )
					return 1;

				AnyKey = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/MR", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseTwoKeys( arg[i] + j, &MoveRightHotKey, &MoveRight2HotKey ) )
					return 1;

				AnyKey = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/DL", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseTwoKeys( arg[i] + j, &DelLeftHotKey, &DelLeft2HotKey ) )
					return 1;

				AnyKey = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/DR", 3 ) ) {
				int j = 3;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( ParseTwoKeys( arg[i] + j, &DelRightHotKey, &DelRight2HotKey ) )
					return 1;

				AnyKey = TRUE;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else {
				DisplayErrorHeader();
				Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Syntax error \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
	}


	if ( Save && Restore ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /R and /S.\r\n" );
		return 1;
	}

	if ( Restore && AnyKey ) {
		DisplayErrorHeader();
		Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /R with a hotkey option.\r\n" );
		return 1;
	}

	if ( List ) {
		if ( Save || Restore || AnyKey ) {
			DisplayErrorHeader();
			Qprintf( GetStdHandle( STD_ERROR_HANDLE ), L"Can\u2019t combine /L with other options.\r\n" );
			return 1;
		}

		ListLegalKeys();
		return 0;
	}

	if ( Save )
		if ( SaveOptions( TRUE ) )
			return 2;

	if ( Restore )
		if ( RestoreOptions( TRUE ) )
			return 2;

	if ( AnyKey )
		QPuts( L"\r\nByArgs hotkeys changed to:\r\n\r\n" );
	else
		QPuts( L"\r\nCurrent ByArgs hotkeys are:\r\n\r\n" );

	ShowKeyAssignments();
	return 0;
}



wchar_t GetEscapeChar()
{
	wchar_t temp[SHORT_BUF_LEN] = L"%@option[escapechar]";
	ExpandVariables( temp, 1 );

	if ( temp[0] && ( temp[1] == '\0' ) )
		return temp[0];
	else
		return '^';
}



/*
wchar_t GetCommandSeparator()
{
	wchar_t temp[SHORT_BUF_LEN] = L"%@option[commandsep]";
	ExpandVariables( temp, 1 );

	if ( temp[0] && ( temp[1] == '\0' ) )
		return temp[0];
	else
		return '&';
}



BOOL GetUnixPaths()
{
	wchar_t temp[SHORT_BUF_LEN] = L"%@option[unixpaths]";
	ExpandVariables( temp, 1 );

	return ( !_wcsicmp( temp, L"Yes" ) ) ? TRUE : FALSE;
}
*/



int FindArgStart( LPTSTR String, int *Cursor, int Flags )
{
	if ( !String )
		return 0;
	else if ( String[0] == '\0' )
		return 0;

	int Direction = Flags & FIRST_AFTER_CURSOR;
	BOOL InArg = FALSE, Quotes = FALSE;
	int ArgStart = -1;
	wchar_t ch = 0x00;
	
	wchar_t Escape		= GetEscapeChar();

	int InBrackets = 0;
	BOOL SaveInArg[MAX_NESTING];
	int SaveQuotes[MAX_NESTING];


	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( iswspace( ch ) ) {
			if ( !Quotes )
				InArg = FALSE;
		}
		else if ( InBrackets && ( ch == ',' ) ) {
			if ( !Quotes )
				InArg = FALSE;
		}
		else {
			if ( !InArg ) {
				InArg = TRUE;

				if ( Direction == LAST_BEFORE_CURSOR ) {
					if ( i < *Cursor )
						ArgStart = i;
				}
				else if ( Direction == FIRST_AFTER_CURSOR ) {
					if ( ( i > *Cursor ) && ( ArgStart < 0 ) )
						ArgStart = i;
				}
			}

			if ( ( ch == Escape ) && String[i + i] )
				i++;
			else if ( ch == QUOTE )
				Quotes = !Quotes;
			else if ( ch == '/' ) {
				if ( !Quotes && !InBrackets ) {
					if ( Direction == LAST_BEFORE_CURSOR ) {
						if ( i < *Cursor )
							ArgStart = i;
					}
					else if ( Direction == FIRST_AFTER_CURSOR ) {
						if ( ( i > *Cursor ) && ( ArgStart < 0 ) )
							ArgStart = i;
					}
				}
			}
			else if ( ( ch == '%' ) && String[i + 1] ) {
				i++;

				if ( String[i] == '@' )
					i++;

				while ( IsVarnameChar( String[i] ) )
					i++;

				if ( String[i] == '[' ) {
					SaveInArg[InBrackets]	= InArg;
					SaveQuotes[InBrackets]	= Quotes;
					Quotes = FALSE;
					InArg = FALSE;

					if ( InBrackets < ( MAX_NESTING - 1 ) )
						InBrackets++;
				}
			}
			else if ( InBrackets && !Quotes && ( ch == ']' ) ) {
				InBrackets--;
				InArg	= SaveInArg[InBrackets];
				Quotes	= SaveQuotes[InBrackets];

				if ( Flags & STOP_AT_CLOSE_BRACKET ) {
					if ( Direction == LAST_BEFORE_CURSOR ) {
						if ( i < *Cursor )
							ArgStart = i;
					}
					else if ( Direction == FIRST_AFTER_CURSOR ) {
						if ( ( i > *Cursor ) && ( ArgStart < 0 ) )
							ArgStart = i;
					}
				}
			}
		}
	}

	if ( ArgStart < 0 )
		return 0;

	*Cursor = ArgStart;
	return 1;
}



int DeleteCharacters( LPTSTR String, int From, int To )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	if ( ( From < 0 ) || ( To < 0 ) )
		return -1;

	if ( From > To ) {
		int temp = From;
		From = To;
		To = temp;
	}

	ClearEditLine( String );

	if ( String[To] == '\0' )
		String[From] = '\0';
	else {
		wchar_t ch = 0x01;

		for ( int i = 0; ch; i++ ) {
			ch = String[To + i + 1];
			String[From + i] = ch;
		}
	}
	
	return 0;
}



DLLExports int WINAPI key( LPKEYINFO24 ki )
{
	if ( ki->nKey == 0 ) {
		if ( !NewKeyHook )
			return 0;
		else if ( ki->pszKey == NULL )
			return 0;
	}


	int inKey = ki->nKey;
	int shiftKeys = 0;

	if ( NewKeyHook ) {
		if ( !inKey && ki->pszKey ) {
			inKey = ParseKeyName24( ki->pszKey );

			if ( !inKey )
				return 0;
		}
	}


	if ( GetAsyncKeyState( VK_SHIFT ) & 0x8000 )
		shiftKeys |= 1;

	if ( GetAsyncKeyState( VK_CONTROL ) & 0x8000 )
		shiftKeys |= 2;

	if ( GetAsyncKeyState( VK_MENU ) & 0x8000 )
		shiftKeys |= 4;

	if ( shiftKeys == 3 )
		inKey |= 0x20000;
	else if ( shiftKeys == 6 )
		inKey |= 0x40000;
	else if ( NewKeyHook && ( shiftKeys == 7 ) )
		inKey |= 0x6000;

	if ( !inKey )
		return 0;


	int Cursor = (int) ( ki->pszCurrent - ki->pszLine );
	int len = (int) wcslen( ki->pszLine );

	if ( ( inKey == MoveLeftHotKey ) || ( inKey == MoveLeft2HotKey ) ) {
		//  MessageBox( GetForegroundWindow(), L"MoveLeftHotKey", PLUGIN_NAME, MB_OK | MB_SETFOREGROUND );

		if ( FindArgStart( ki->pszLine, &Cursor, LAST_BEFORE_CURSOR ) ) {
			ki->pszCurrent = ki->pszLine + Cursor;
			ki->fRedraw = 1;
		}
		else if ( Cursor ) {
			ki->pszCurrent = ki->pszLine;
			ki->fRedraw = 1;
		}
		ki->nKey = 0;
		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
	}
	else if ( ( inKey == MoveRightHotKey ) || ( inKey == MoveRight2HotKey ) ) {
		//  MessageBox( GetForegroundWindow(), L"MoveRightHotKey", PLUGIN_NAME, MB_OK | MB_SETFOREGROUND );

		if ( FindArgStart( ki->pszLine, &Cursor, FIRST_AFTER_CURSOR ) ) {
			ki->pszCurrent = ki->pszLine + Cursor;
			ki->fRedraw = 1;
		}
		else if ( Cursor < len ) {
			ki->pszCurrent = ki->pszLine + len;
			ki->fRedraw = 1;
		}
		ki->nKey = 0;
		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
}
	else if ( ( inKey == DelLeftHotKey ) || ( inKey == DelLeft2HotKey ) ) {
		//  MessageBox( GetForegroundWindow(), L"DelLeftHotKey", PLUGIN_NAME, MB_OK | MB_SETFOREGROUND );

		if ( FindArgStart( ki->pszLine, &Cursor, LAST_BEFORE_CURSOR ) ) {
			DeleteCharacters( ki->pszLine, Cursor, (int) ( ki->pszCurrent - ki->pszLine - 1 ) );
			ki->pszCurrent = ki->pszLine + Cursor;
			ki->fRedraw = 1;
		}
		else if ( Cursor ) {
			DeleteCharacters( ki->pszLine, 0, Cursor - 1 );
			ki->pszCurrent = ki->pszLine;
			ki->fRedraw = 1;
		}
		ki->nKey = 0;
		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
	}
	else if ( ( inKey == DelRightHotKey ) || ( inKey == DelRight2HotKey ) ) {
		// MessageBox( GetForegroundWindow(), L"DelRightHotKey", PLUGIN_NAME, MB_OK | MB_SETFOREGROUND );

		if ( FindArgStart( ki->pszLine, &Cursor, FIRST_AFTER_CURSOR | STOP_AT_CLOSE_BRACKET ) ) {
			DeleteCharacters( ki->pszLine, (int) ( ki->pszCurrent - ki->pszLine ), Cursor - 1 );
			ki->fRedraw = 1;
		}
		else if ( Cursor < len ) {
			DeleteCharacters( ki->pszLine, Cursor, len );
			ki->fRedraw = 1;
		}
		ki->nKey = 0;
		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
	}
	/*
	else if ( inKey == 0x1000e ) {
		CONSOLE_SELECTION_INFO csi;
		csi.dwFlags = 0;
		csi.srSelection.Bottom = 0;
		csi.srSelection.Top = 0;
		csi.srSelection.Left = 0;
		csi.srSelection.Right = 0;

		GetConsoleSelectionInfo( &csi );
		
		FmtMsgBox( 0, PLUGIN_NAME, L"csi.srSelection = %d, %d, %d, %d", csi.srSelection.Bottom, csi.srSelection.Left, csi.srSelection.Top, csi.srSelection.Right );

		ki->nKey = 0;
		if ( NewKeyHook )
			ki->pszKey[0] = '\0';
	}
	*/

	return 0;
}



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	if ( !_wcsicmp( inString, L"byargs" ) ) {
		MessageBox( GetForegroundWindow(), ByArgsHelpText, HelpWindowTitle, MB_OK );
		return 1;
	}
	else
		return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( !_wcsicmp( inString, L"byargs" ) ) {
		wcscpy_s( outString, SHORT_BUF_LEN, UsageText( ByArgsHelpText ) );
		return 1;
	}
	else
		return 0;
}
